#include <queue>
#include <stack>
#include <iterator>
#include <random>
#include <cmath>
#include <algorithm>
#include "graph.hpp"


uint32_t adj_matrix::new_node(){
	//TODO: should the diagonal contain a 0 weight edge?
	while(matrix.find(++last_node) != matrix.end());
	matrix[last_node] = unordered_map<uint32_t,uint32_t>();
	for(auto &[key, val] : matrix){
		if(key > last_node) val[last_node] = 0;
		else matrix[last_node][key] = 0;
	}
	return last_node;
}


void adj_matrix::rem_node(uint32_t index){
	matrix.erase(index);
	for(auto &[key, val] : matrix){
		if(key > index) val.erase(index);
	}
}


uint32_t adj_matrix::add_edge(uint32_t head, uint32_t tail, uint32_t weight){
	if(head < tail) swap(head, tail);
	if(matrix[head][tail] == 0){
		while(edge_weights.find(++last_edge) != edge_weights.end());
		matrix[head][tail] = last_edge;
		edge_weights[last_edge] = unordered_multiset<uint32_t>();
	}
	edge_weights[last_edge].insert(weight);
	return matrix[head][tail];
}


void adj_matrix::rem_edge(uint32_t head, uint32_t tail){
	if(head < tail) swap(head, tail);
	uint32_t edge = matrix[head][tail];
	if(edge == 0) return;
	matrix[head][tail] = 0;
	edge_weights.erase(edge);
}


// void adj_matrix::new_weight(uint32_t edge, uint32_t weight){
// 	if(edge == 0) return;
// 	edge_weights[edge] = weight;
// }


uint32_t adj_matrix::get_edge(uint32_t head, uint32_t tail){
	if(head < tail) swap(head, tail);
	return matrix[head][tail];
}


unordered_set<uint32_t> adj_matrix::get_edges(uint32_t node_index){
	unordered_set<uint32_t> output;
	for(auto &[key, val] : matrix){
		if(key > node_index && val[node_index]) output.insert(val[node_index]);
		else if(matrix[node_index][key]) output.insert(matrix[node_index][key]);
	}
	return output;
}


unordered_multiset<uint32_t> adj_matrix::get_weights(uint32_t edge_index){
	return edge_weights[edge_index];
}

vector<uint32_t> adj_matrix::get_adj(uint32_t node){
	vector<uint32_t> output;
	for(auto &[k, v] : matrix){
		if(v.find(node) != v.end() && v[node]) 
			output.push_back(k);
	}
	for(auto &[k, v] : matrix[node]){
		if(v) output.push_back(k);
	}
	//A non-deterministic  ordering of the neighbor nodes will result 
	//in different paths taken and should make infinite loops 
	//impossible as long as the target node is reachable
	random_shuffle(output.begin(), output.end());
	return output;
}


bool adj_matrix::is_adjacent(uint32_t head, uint32_t tail){
	if(head < tail) swap(head, tail);
	return (matrix[head][tail]);
}


bool adj_matrix::check_coloring(unordered_map<uint32_t,uint32_t> coloring){
	for(auto &[node, color] : coloring){
		for(auto adj : get_adj(node)){
			if(coloring[adj] == color) return false;
		}
	}
	return true;
}


unordered_map<uint32_t,uint32_t>* adj_matrix::k_color(uint32_t k){
	auto output = new unordered_map<uint32_t,uint32_t>;
	vector<uint32_t> keys; 
	for(auto &[k, v] : matrix) {
		keys.push_back(k);
		(*output)[k] = 0;
	}
	for(uint64_t i = 0; i < pow(k, matrix.size()); i++){
		bool carry = true;
		for(uint32_t j = 0; j < keys.size() && carry; j++){
			(*output)[keys[j]] = ((*output)[keys[j]]+1)%k;
			if(!((*output)[keys[j]])) carry = false;
		}
		if(check_coloring(*output)) return output;
	}
	return NULL;
}


unordered_map<uint32_t,uint32_t>* adj_matrix::minimal_coloring(uint32_t max){
	for(uint32_t i = 2; i <= max; i++){
		auto output = k_color(i);
		if(output) return output;
	}
	return NULL;
}


void adj_matrix::print_node(uint32_t node){}


void adj_matrix::print(){
	//Formatting is only pretty for one or two digit nodes and weights
	vector<uint32_t> keys;
	for(auto &[key, _] : matrix) keys.push_back(key);
	sort(keys.begin(), keys.end());

	//Label Columns
	cout << "    ";
	for(auto k : keys) printf("%2i ", k);
	cout << endl;

	//Label and fill rows
	for(auto k : keys){
		printf("%2i: ", k);
		for(auto &[_, val] : matrix[k]){
			if(val == 0) printf(" . ");
			else printf("%2i ", *(edge_weights[val].begin()));
		} cout << endl;
	} cout << endl;

	//Print adjacencies
	// for(auto &[k,_] : matrix){
	// 	cout << k << ": ";
	// 	for(auto i : get_adj(k)) cout << i << " ";
	// 	cout << endl;
	// }
}


/*****************\
| Graph Functions |
\*****************/


//Breadth First Search
vector<uint32_t> Graph::BFS(uint32_t s, uint32_t e){
	unordered_map<uint32_t,uint32_t> visited = {{s, s}};
	queue<uint32_t> frontier;
	frontier.push(s);

	//Perform the search
	while(visited.find(e) == visited.end()){
		if(frontier.size() == 0) return vector<uint32_t>();
		uint32_t node = frontier.front();
		frontier.pop();
		for(uint32_t neighbor : get_adj(node)){
			if(visited.find(neighbor) != visited.end()) continue;
			visited[neighbor] = node;
			frontier.push(neighbor);
		}
	}

	//Build the path from the end to start using the visited map
	vector<uint32_t> output;
	output.push_back(e);
	for(uint32_t prev = e; prev != s;){
		prev = visited[prev];
		output.push_back(prev);
	}
	reverse(output.begin(), output.end());
	return output;
}


//Depth First Search
vector<uint32_t> Graph::DFS(uint32_t s, uint32_t e, uint32_t max_depth){
	if(s == e) return vector<uint32_t> {s};
	if(max_depth == 0) return vector<uint32_t>();
	
	vector<uint32_t> path = {s};
	stack<tuple<uint32_t,uint32_t>> frontier;
	frontier.push({s, 0});

	while(path[path.size()-1] != e){
		if(frontier.size() == 0) return vector<uint32_t>();
		auto [node, depth] = frontier.top(); frontier.pop();
		while(path.size()-1 > depth) path.pop_back();
		if(depth > 0) path.push_back(node);
		
		for(uint32_t neighbor : get_adj(node)){
			if(neighbor == e){ path.push_back(neighbor); break; }
			if(depth < max_depth-1) 
				frontier.push({neighbor, depth+1});
		}
	}

	//TODO: fix the logic error in the above code that makes this necessary
	//Paths that exceed max_depth aren't properly pruned from path
	vector<uint32_t> fixed_path = {e};
	for(int i = path.size()-2; i >= 0; i--){
		if(is_adjacent(path[i], path[i+1])) fixed_path.push_back(path[i]);
	} reverse(fixed_path.begin(), fixed_path.end());

	return fixed_path;
}


//Iterative Deepening Depth First Search
vector<uint32_t> Graph::IDS(uint32_t s, uint32_t e){
	vector<uint32_t> path;
	for(int i = 1; path.size() == 0; path = DFS(s, e, i++));
	return path;
}